from django.urls import reverse
from django.conf import settings
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.templatetags.static import static
from django.utils.translation import gettext_lazy as _
from wagtail.models import (
    WorkflowMixin,
)
from wagtail.admin.userbar import (
    BaseItem,
    AddPageItem,
    ExplorePageItem,
    EditPageItem,
)
from wagtail import hooks
from ..toolbar import (
    FeditToolbarComponent,
)
from ..utils import (
    FEDIT_PREVIEW_VAR,
    FeditPermissionCheck,
    access_userbar_model,
    is_draft_capable,
    user_can_publish,
    user_can_unpublish,
    user_can_submit_for_moderation,
)

class FeditableModelComponent(FeditToolbarComponent):
    template_name = "wagtail_fedit/userbar/publish/action_button.html"
    action_icon = None
    action_text = None

    def __init__(self, request, instance):
        self.request = request
        self.instance = instance

    def get_context_data(self):
        return super().get_context_data() | {
            "hidden": not self.instance.has_unpublished_changes,
            "action_icon": self.action_icon,
            "action_text": self.action_text,
            "action_url": reverse(
                self.action_url,
                args=[self.instance.pk, self.instance._meta.app_label, self.instance._meta.model_name],
            ),
        }

class UserBarActionPublishComponent(FeditableModelComponent):
    action_url = "wagtail_fedit:publish"
    action_icon = "fedit-eye-open"
    action_text = _("Publish")
    
    def is_shown(self):
        if not super().is_shown():
            return False
        
        return user_can_publish(self.instance, self.request.user, check_for_changes=False)\
               and self.instance.has_unpublished_changes

class UserBarActionUnpublishComponent(FeditableModelComponent):
    action_url = "wagtail_fedit:unpublish"
    action_icon = "fedit-eye-closed"
    action_text = _("Unpublish")
        
    def is_shown(self):
        if not super().is_shown():
            return False
        
        return user_can_unpublish(self.instance, self.request.user)

class UserBarActionSubmitComponent(FeditableModelComponent):
    action_url = "wagtail_fedit:submit"
    action_icon = "fedit-check-list"
    action_text = _("Submit for moderation")

    def is_shown(self):
        if not super().is_shown():
            return False
        
        return user_can_submit_for_moderation(self.instance, self.request.user, check_for_changes=False)\
               and self.instance.has_unpublished_changes

class UserBarActionCancelComponent(FeditableModelComponent):
    action_url = "wagtail_fedit:cancel"
    action_icon = "fedit-stop-sign"
    action_text = _("Cancel Workflow")

    def is_shown(self):
        if not super().is_shown():
            return False
        
        if not is_draft_capable(self.instance):
            return False
        
        if not isinstance(self.instance, WorkflowMixin):
            return False
        
        if not getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
            return False
        
        wf_state = self.instance.current_workflow_state or\
            self.instance.workflow_states.order_by("created_at").last()

        return wf_state and\
            wf_state.status in\
                (wf_state.STATUS_IN_PROGRESS, wf_state.STATUS_NEEDS_CHANGES)

        # workflow_state = self.instance.current_workflow_state
        # return workflow_state and (
        #     workflow_state.status == workflow_state.STATUS_IN_PROGRESS or\
        #     workflow_state.status == workflow_state.STATUS_NEEDS_CHANGES
        # )

class BaseWagtailFeditItem(BaseItem, FeditPermissionCheck):
    def __init__(self, model):
        self.model = model

    def get_context_data(self, request):
        context = super().get_context_data(request)
        context["model"] = self.model
        context["edit_url"] = reverse(
            "wagtail_fedit:editable",
            args=[
                self.model.pk,
                self.model._meta.app_label,
                self.model._meta.model_name,
            ],
        )
        return context

    def render(self, request):
        if not self.has_perms(request, self.model):
            return ""

        return super().render(request)


class WagtailFeditItem(BaseWagtailFeditItem):
    template = "wagtail_fedit/userbar/item_fedit.html"

class WagtailFeditViewLiveItem(BaseWagtailFeditItem):
    template = "wagtail_fedit/userbar/item_fedit_view_live.html"

    def get_context_data(self, request):
        context = super().get_context_data(request)
        
        if hasattr(self.model, "get_url"):
            context["live_url"] = self.model.get_url(request)
        elif hasattr(self.model, "get_absolute_url"):
            context["live_url"] = self.model.get_absolute_url()

        return context
    
class WagtailFeditPublishItem(BaseWagtailFeditItem):
    template = "wagtail_fedit/userbar/publish/item_fedit_publishing.html"

    def render(self, request):

        self.can_publish = user_can_publish(self.model, request.user, check_for_changes=False)
        self.can_unpublish = user_can_unpublish(self.model, request.user)
        self.can_submit_for_moderation = user_can_submit_for_moderation(
            self.model, request.user, check_for_changes=False,
        )

        if not self.can_publish\
                and not self.can_unpublish\
                and not self.can_submit_for_moderation:
            return ""

        return super().render(request)

    def get_context_data(self, request):

        buttons = [
            UserBarActionPublishComponent(request, self.model),
            UserBarActionSubmitComponent(request, self.model),
            UserBarActionUnpublishComponent(request, self.model),
            UserBarActionCancelComponent(request, self.model),
        ]

        return super().get_context_data(request) | {
            "buttons": list(
                filter(None, map(lambda x: x.render(), buttons))
            ),
        }

def retrieve_page_model(items):
    # Retrieve page from other items...
    # sad we are not just provided with the context or page too.
    for item in items:
        if isinstance(item, (AddPageItem, ExplorePageItem, EditPageItem))\
                or hasattr(item, "page"):
            
            return item.page
    return None

@hooks.register("insert_global_admin_css")
def fedit_admin_js():
    return mark_safe(format_html(
        '<link rel="stylesheet" href="{}">',
        static("wagtail_fedit/css/userbar-menu.css"),
    ))

@hooks.register("construct_wagtail_userbar")
def add_fedit_userbar_item(request, items):
    model = access_userbar_model(request)

    if not model:
        model = retrieve_page_model(items)

    local_items = []
    
    if getattr(request, FEDIT_PREVIEW_VAR, False):
        if is_draft_capable(model):
            local_items.append(
                WagtailFeditPublishItem(model),
            )

        if hasattr(model, "get_absolute_url") or hasattr(model, "get_url"):
            local_items.append(
                WagtailFeditViewLiveItem(model),
            )

    if model and not getattr(request, FEDIT_PREVIEW_VAR, False):
        local_items.append(
            WagtailFeditItem(model),
        )

    items[:] = local_items + items







from typing import Type
from django import forms
from django.db import models
from django.utils.translation import gettext as _
from django.http import (
    HttpRequest,
)
from wagtail.admin.forms import (
    WagtailAdminModelForm,
)
from wagtail.models import (
    RevisionMixin,
)

from ..hooks import (
    REGISTER_FIELD_WIDGETS,
)

from wagtail import hooks



_looked_for_widgets = False
_widgets = {}


def _look_for_widgets():
    global _looked_for_widgets
    if not _looked_for_widgets:
        _looked_for_widgets = True
        for fn in hooks.get_hooks(REGISTER_FIELD_WIDGETS):
            _widgets.update(fn(_widgets))


def get_widget_for_field(field: models.Field) -> Type[forms.Widget]:
    """
    Return a widget for a field.
    """
    _look_for_widgets()
    global _widgets

    if isinstance(field, models.ForeignKey):
        field = field.related_model
        widget = _widgets.get(field)
    else:
        widget = _widgets.get(field.__class__)

    if widget is None:
        pass

    return widget


def save_possible_revision(instance: models.Model, request: HttpRequest, **kwargs) -> models.Model:
    """
    Save an instance as a revision if the model supports it.
    """
    if isinstance(instance, RevisionMixin):
        instance = instance.save_revision(
            user=request.user,
            **kwargs,
        )
        instance = instance.as_object()
    else:
        instance.save()

    return instance


def get_form_class_for_fields(form_model: models.Model, form_fields: list[str]) -> Type["PossibleRevisionForm"]:
    """
    Return a form class for a model with specific fields.
    This is similar to django's modelform_factory, but with the added benefit of using the custom widgets.
    It also keeps the revision functionality in mind.
    """

    if hasattr(form_model, "get_fedit_form"):
        return form_model.get_fedit_form(form_fields)
    
    if form_fields == "__all__" or tuple(form_fields) == ("__all__", ):
        form_fields = [f.name for f in form_model._meta.fields]
    
    form_widgets = {}
    for field_name in form_fields:
        field = form_model._meta.get_field(field_name)
        widget = get_widget_for_field(field)
        if widget:
            form_widgets[field_name] = widget

    class RevisionForm(PossibleRevisionForm):
        class Meta:
            model = form_model
            fields = form_fields
            widgets = form_widgets

    return RevisionForm


class PossibleRevisionFormMixin:
    """
    A form that can save a revision if the model is a RevisionMixin.
    Otherwise resorts to the default save method; this saves the (live) instance.
    """
    def __init__(self, *args, request = None, **kwargs):
        self.request = request
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        self.instance = super().save(commit=False)
        if commit:
            self.instance = save_possible_revision(
                self.instance,
                self.request,
            )
        return self.instance


class PossibleRevisionForm(PossibleRevisionFormMixin, WagtailAdminModelForm):
    pass



from django.test import TestCase
from django.db import models
from django.urls import reverse
from django.test import RequestFactory
from django.contrib.auth.models import (
    User,
    Permission,
    AnonymousUser,
)
from wagtail.models import (
    Page,
)
from ..models import (
    BasicModel,
    EditablePageModel,
    EditableFullModel,
    EditableDraftModel,
    EditableRevisionModel,
    EditablePreviewModel,
    EditableLockModel,
)

import urllib.parse

from wagtail_fedit.adapters import BaseAdapter

TEST_BLOCK_DATA = [
    {
        "type": "heading_component",
        "value": {
            "heading": "This is pretty cool!!!",
            "subheading": "RIGHT?! FUCK YES!"
        },
        "id": "0bc9f67e-f116-48a6-9ca1-6c11d39b54e8"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "AWESOME!!",
            "subheading": "RIGHT?!"
        },
        "id": "d543a6bf-34dc-4365-a3fa-d302561930ae"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "WORKS NOW!",
            "subheading": "!!!!!!!!!!!!1"
        },
        "id": "c49abcae-3c66-4fc7-979d-35407226b9f5"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "Heading!!!!",
            "subheading": "Subheading"
        },
        "id": "7bd7bc3a-1d2d-4182-8726-b257beace968"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "Hey!",
            "subheading": "Subheading!"
        },
        "id": "74a94baa-acf4-49ab-be5f-9c8a70cbc623"
    },
    {
        "type": "flat_menu_component",
        "value": {
            "title": "Test Title123123! HAHA!",
            "subtitle": "<p data-block-key=\"306j3\">i am so<b><i> happy</i></b></p>",
            "items": [
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 1"
                        }
                    },
                    "id": "c757f54d-0df5-4b35-8a06-4174f180ec41"
                },
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 2"
                        }
                    },
                    "id": "ec3d73d1-fd01-49ba-840a-d44586ac0025"
                },
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 3"
                        }
                    },
                    "id": "a98a19c6-2ead-4e69-9ea2-3158c7e82976"
                },
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 4"
                        }
                    },
                    "id": "db7183a2-d9dd-4fbd-9e42-fd2b9ebf0458"
                }
            ]
        },
        "id": "3e9144fd-5fa5-47f8-917e-8fe87c15da01"
    }
]



class BaseFEditTest(TestCase):
    # Block ID for a sub-block present in test data.
    BLOCK_ID = "c757f54d-0df5-4b35-8a06-4174f180ec41"
    
    def setUp(self):
        super().setUp()

        self.request_factory = RequestFactory()
        root_page: Page = Page.add_root(title="Root Page")
        self.page_model = root_page.add_child(
            instance=EditablePageModel(
                title="Editable Page",
                body="Editable Page Body",
                content=TEST_BLOCK_DATA,
                url_path="/home/editable-page/",
            )
        )
        self.full_model = EditableFullModel.objects.create(
            title="Full Model",
            body="Full Model Body",
            content=TEST_BLOCK_DATA,
        )
        self.draft_model = EditableDraftModel.objects.create(
            title="Draft Model",
            body="Draft Model Body",
            content=TEST_BLOCK_DATA,
        )
        self.revision_model = EditableRevisionModel.objects.create(
            title="Revision Model",
            body="Revision Model Body",
            content=TEST_BLOCK_DATA,
        )
        self.preview_model = EditablePreviewModel.objects.create(
            title="Preview Model",
            body="Preview Model Body",
            content=TEST_BLOCK_DATA,
        )
        self.basic_model = BasicModel.objects.create(
            title="Basic Model",
            body="Basic Model Body",
            content=TEST_BLOCK_DATA,
        )

        # Additional models for other functionality
        self.admin_user = User.objects.create_superuser(
            username="admin",
            email="admin@localhost",
            password="admin"
        )
        self.other_admin_user = User.objects.create_superuser(
            username="other_admin",
            email="other_admin@localhost",
            password="other_admin"
        )
        self.regular_user = User.objects.create_user(
            username="regular",
            email="regular@localhost",
            password="regular"
        )
        self.anonymous_user = AnonymousUser()
        self.regular_user.user_permissions.add(Permission.objects.get(
            codename="access_admin",
            content_type__app_label="wagtailadmin"
        ))

        self.lock_model = EditableLockModel.objects.create(
            title="Lock Model",
            body="Lock Model Body",
            content=TEST_BLOCK_DATA,
            locked=True,
            locked_by=self.admin_user,
            has_unpublished_changes=True,
        )

        self.basic_model.related_field = self.basic_model.__class__.objects.create(
            title="Related Field",
            body="Related Field Body",
            content=TEST_BLOCK_DATA,
        )

        self.basic_model.save()
        
        self.models: list[models.Model] = [
            self.full_model,
            self.draft_model,
            self.revision_model,
            self.preview_model,
            self.basic_model,
        ]

    def get_editable_url(self, object_id, app_label, model_name):
        url_name = "editable"
        return reverse(
            f"wagtail_fedit:{url_name}",
            kwargs={
                "object_id": object_id,
                "app_label": app_label,
                "model_name": model_name
            }
        )

    def get_url_for(self, url_name, app_label, model_name, model_id):
        return reverse(
            f"wagtail_fedit:{url_name}",
            kwargs={
                "object_id": model_id,
                "app_label": app_label,
                "model_name": model_name,
            }
        )

    def get_field_url(self, field_name, app_label, model_name, model_id):
        url_name = "edit"
        return reverse(
            f"wagtail_fedit:{url_name}",
            kwargs={
                "adapter_id": "field",
                "field_name": field_name,
                "app_label": app_label,
                "model_name": model_name,
                "model_id": model_id
            }
        )
    
    def get_model_url(self, app_label, model_name, model_id):
        url_name = "edit"
        return reverse(
            f"wagtail_fedit:{url_name}",
            kwargs={
                "adapter_id": "model",
                "app_label": app_label,
                "model_name": model_name,
                "model_id": model_id
            }
        )


    def get_block_url(self, block_id, field_name, app_label, model_name, model_id):
        url_name = "edit"
        url = reverse(
            f"wagtail_fedit:{url_name}",
            kwargs={
                "adapter_id": "block",
                "field_name": field_name,
                "app_label": app_label,
                "model_name": model_name,
                "model_id": model_id
            }
        )
        adapter = BaseAdapter(
            BasicModel(), "title", None, block_id=block_id,
        )
        encoded = adapter.encode_shared_context()
        encoded = urllib.parse.urlencode({"shared_context": encoded})
        return f"{url}?{encoded}"





from django.db import models
from django.http import HttpResponse
from django.utils.text import slugify
from wagtail.snippets.models import register_snippet
from django.template import Template, Context
from wagtail import blocks
from wagtail.admin.panels import (
    FieldPanel,
)
from wagtail.fields import (
    StreamField,
    StreamValue,
)
from wagtail.models import (
    Page,
    LockableMixin,
    RevisionMixin,
    PreviewableMixin,
    WorkflowMixin,
    DraftStateMixin,
)
from wagtail_fedit.models import (
    FEditableMixin,
    FeditPermissionTester,
    ModelPermissionPolicy,
)

class HeadingComponent(blocks.StructBlock):
    heading = blocks.CharBlock(max_length=25)
    subheading = blocks.CharBlock(max_length=40)

    def render(self, value, context=None):
        template = Template("""
            <h1>{{ heading }}</h1>
            <h2>{{ subheading }}</h2>
        """)

        context = self.get_context(value, parent_context=context)
        return template.render(Context(context))

class LinkBlock(blocks.StructBlock):
    text = blocks.CharBlock(max_length=25)

    def render(self, value, context=None):
        template = Template("""
            <a href="#">{{ text }}</a>
        """)

        context = self.get_context(value, parent_context=context)
        return template.render(Context(context))

class MenuItemBlock(blocks.StructBlock):
    link = LinkBlock()

    def render(self, value, context=None):
        template = Template("""
            {% load wagtailcore_tags %}
            <li>{% include_block self.link %}</li>
        """)

        context = self.get_context(value, parent_context=context)
        return template.render(Context(context))

class FlatMenuComponent(blocks.StructBlock):
    title = blocks.CharBlock(max_length=25)
    subtitle = blocks.RichTextBlock()
    items = blocks.ListBlock(
        MenuItemBlock()
    )

    def render(self, value, context=None):
        template = Template("""
            {% load wagtailcore_tags %}
            <h1>{{ title }}</h1>
            <p>{{ subtitle }}</p>
            <ul>
                {% for item in items %}
                    {% include_block item %}
                {% endfor %}
            </ul>
        """)

        context = self.get_context(value, parent_context=context)
        return template.render(Context(context))


class BaseEditableMixin:
    def get_url(self, request):
        return f"/{slugify(self.title)}/"
    
    def serve(self, request, *args, **kwargs):
        return HttpResponse(self.body)

    def render_as_content(self, request, context=None):
        return f"<h1>{self.title}</h1><p>{self.body}</p>"

@register_snippet
class BasicModel(models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)
    related_field = models.ForeignKey("self", on_delete=models.CASCADE, null=True, blank=True)

    panels = [
        FieldPanel("title"),
        FieldPanel("body"),
    ]

    def render_as_content(self, request, context=None):
        return f"<h1>{self.title}</h1><p>{self.body}</p>"


@register_snippet
class EditableFullModel(BaseEditableMixin, FEditableMixin):
    title = models.CharField(max_length=255)
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)


@register_snippet
class EditableDraftModel(BaseEditableMixin, DraftStateMixin, RevisionMixin, models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)


@register_snippet
class EditableRevisionModel(BaseEditableMixin, RevisionMixin, models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)


@register_snippet
class EditablePreviewModel(BaseEditableMixin, PreviewableMixin, models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)

@register_snippet
class EditableLockModel(BaseEditableMixin, WorkflowMixin, DraftStateMixin, RevisionMixin, LockableMixin, models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)

    panels = [
        FieldPanel("title"),
        FieldPanel("body"),
    ]
    
    def get_permissions_policy(self):
        return ModelPermissionPolicy(self.__class__)
    
    def permissions_for_user(self, user):
        return FeditPermissionTester(
            self,
            user=user,
            policy=self.get_permissions_policy()
        )


class EditablePageModel(Page):
    body = models.TextField()
    content: StreamValue = StreamField([
        ("heading_component", HeadingComponent()),
        ("flat_menu_component", FlatMenuComponent())
    ], use_json_field=True)

    content_panels = Page.content_panels + [
        FieldPanel("body")
    ]

    promote_panels = []
    settings_panels = []

    def render_as_content(self, request, context=None):
        return f"<h1>{self.title}</h1><p>{self.body}</p>"


from django.utils.functional import classproperty
from .base import Keyword
from .block import BlockAdapter
from .field import FieldAdapter
from itertools import chain


# https://stackoverflow.com/a/29088221/18020941
def dj_model_to_dict(instance):
    opts = instance._meta
    data = {}
    for f in chain(opts.concrete_fields, opts.private_fields):
        data[f.name] = f.value_from_object(instance)
    for f in opts.many_to_many:
        data[f.name] = [i.id for i in f.value_from_object(instance)]
    return data


class FuncAdapterMixin:
    base_identifier = None
    base_keywords = ()
    js_constructor = "wagtail_fedit.editors.BaseFuncEditor"
    js_function = None

    @classproperty
    def identifier(cls):
        if not cls.base_identifier:
            raise NotImplementedError("base_identifier is required")
        
        return f"{cls.base_identifier}_func"
    
    def get_element_id_parts(self) -> list[str]:
        return super().get_element_id_parts() + [
            self.kwargs["target"],
        ]

    @classproperty
    def keywords(cls):

        if cls.js_function:
            return cls.base_keywords + (
                Keyword(
                    "target",
                    help_text="The target element to apply the function to.",
                    type_hint="str",
                ),
            )

        return cls.base_keywords + (
            Keyword(
                "name",
                help_text="The name of the javascript function to call.",
                type_hint="str",
            ),
            Keyword(
                "target",
                help_text="The target element to apply the function to.",
                type_hint="str",
            ),
        )

    def get_response_data(self, parent_context=None):
        data = super().get_response_data(parent_context)
        name = self.js_function or self.kwargs["name"]
        target = self.kwargs["target"]
        target = target.format(object=self.object)
        data["func"] = {
            "name": name,
            "target": target,
        }
        return data

class BaseFieldFuncAdapter(FuncAdapterMixin, FieldAdapter):
    base_identifier = FieldAdapter.identifier
    base_keywords = FieldAdapter.keywords

class BaseBlockFuncAdapter(FuncAdapterMixin, BlockAdapter):
    base_identifier = BlockAdapter.identifier
    base_keywords = BlockAdapter.keywords

    def get_response_data(self, parent_context=None):
        data = super().get_response_data(parent_context)
        data["block"] = self.block.get_prep_value()
        return data




from django.conf import settings



SIGN_SHARED_CONTEXT = getattr(settings, "WAGTAIL_FEDIT_SIGN_SHARED_CONTEXT", True)
"""
Sign the shared context with a secret key.
This is useful to prevent tampering with the shared context.
It will also be compressed with zlib if available.
It might not be in your site's security model to need this.
"""


SHARE_WITH_SESSIONS = getattr(settings, "WAGTAIL_FEDIT_SHARE_WITH_SESSIONS", False)
"""
Share the context with the session.
This is useful if you are running into limits with the URL length.
This will store the context in the session and pass the session
key to the iFrame instead of the context.
"""

USE_ADAPTER_SESSION_ID = getattr(settings, "WAGTAIL_FEDIT_USE_ADAPTER_SESSION_ID", True)
"""
Use the get_element_id method of the adapter to generate a session ID.
This is useful to not clutter session data too much.
If `False`, the session ID will be generated each time the page is loaded.
"""

TRACK_LOCALES = getattr(settings, "WAGTAIL_FEDIT_TRACK_LOCALES", False)
"""
Track the locales of the users across the views.
**This sets the initial request.LANGUAGE_CODE (if available) in the shared context.**
If this is false, there is no guarantee that the language of a saved field/model
will be the same as it initially was, generally it will be - however this might mess up with Wagtail's `Page.locale` and
the page not being available in the context afterwards.
"""

TIPPY_ENABLED = getattr(settings, "WAGTAIL_FEDIT_TIPPY_ENABLED", True)
"""
Enable Tippy.js tooltips for toolbar buttons.
"""




from .blocks import (
    get_block_form_class,
    BlockEditForm,
)
from .fields import (
    PossibleRevisionFormMixin,
    PossibleRevisionForm,
    get_form_class_for_fields,
    get_widget_for_field,
)


import uuid
from django.db import models
from django.utils import translation
from django.utils.translation import gettext_lazy as _
from django.utils.safestring import mark_safe
from django.http import HttpRequest

from wagtail.log_actions import log
from wagtail.models import RevisionMixin
from wagtail.fields import StreamField
from wagtail import hooks

from .base import (
    BlockFieldReplacementAdapter,
    VARIABLES,
)
from ..hooks import (
    FIELD_EDITOR_SIZE,
)
from ..utils import (
    use_related_form,
    model_diff,
    get_model_string,
    get_field_content,
    is_draft_capable,
    FeditIFrameMixin,
)
from ..forms import (
    fields as field_forms,
)



class FieldAdapter(BlockFieldReplacementAdapter):
    """
    An adapter for editing any model field.
    This will render the field and replace it on the frontend
    on successful form submission.
    """
    identifier = "field"

    def __init__(self, object, field_name: str, request: HttpRequest, **kwargs):
        super().__init__(object, field_name, request, **kwargs)

        self.original_object = object
        self.initial_field_value = self.field_value

        if self.meta_field.is_relation and use_related_form(self.meta_field):
            # If the field is a model, we want to edit the model itself
            # We can do this by getting the fields from the model
            self.object = self.field_value
            self.form_class = field_forms.get_form_class_for_fields(
                self.meta_field.related_model,
                getattr(
                    self.meta_field.related_model,
                    "fedit_fields", "__all__"
                ),
            )
        else:
            self.form_class = field_forms.get_form_class_for_fields(
                self.model,
                [field_name],
            )


    def get_header_title(self):
        meta_field: models.Field = self.model._meta.get_field(self.field_name)

        model_string = get_model_string(self.original_object)
        if self.original_object != self.object:
            instance_string = get_model_string(self.object)
            return _("Edit model '%(instance_string)s' for %(model_name)s '%(model_string)s'") % {
                "instance_string": instance_string,
                "model_name": self.model._meta.verbose_name,
                "model_string": model_string,
            }

        return _("Edit field '%(field_name)s' for %(model_name)s '%(model_string)s'") % {
            "field_name": meta_field.verbose_name,
            "model_name": self.model._meta.verbose_name,
            "model_string": model_string,
        }
    
    def get_help_text(self):
        if is_draft_capable(self.original_object)\
                and is_draft_capable(self.object)\
                and model_diff(self.object, self.original_object):
            return {
                "status": "warning",
                "heading": FeditIFrameMixin.HEADING_SUPPORTS_DRAFTS,
                "title": FeditIFrameMixin.TITLE_SUPPORTS_DRAFTS,
                "text": mark_safe(_("You must publish %(model)s and the related object of type %(related_verbose_name)s (%(related_model)s) to make any changes visible.") % {
                    "model": get_model_string(self.original_object, publish_url=True, request=self.request),
                    "related_verbose_name": self.object._meta.verbose_name,
                    "related_model": get_model_string(self.object, publish_url=True, request=self.request),
                })
            }

        elif is_draft_capable(self.original_object)\
                and not is_draft_capable(self.object)\
                and model_diff(self.object, self.original_object):
            return {
                "status": "warning",
                "heading": FeditIFrameMixin.HEADING_SUPPORTS_DRAFTS,
                "title": FeditIFrameMixin.TITLE_SUPPORTS_DRAFTS,
                "text": mark_safe(FeditIFrameMixin.TEXT_PUBLISH_DRAFTS % {
                    "model": get_model_string(self.original_object, publish_url=True, request=self.request),
                })
            }
        
        elif not is_draft_capable(self.original_object)\
                and is_draft_capable(self.object)\
                and model_diff(self.object, self.original_object):
            return {
                "status": "warning",
                "heading": _("Publishing related object required."),
                "title": _("The object you are editing supports drafts."),
                "text": mark_safe(_("You must publish the related object of type %(type)s (%(model)s) to make any changes visible.") % {
                    "type": self.original_object._meta.verbose_name,
                    "model": get_model_string(self.original_object, publish_url=True, request=self.request),
                })
            }

        return super().get_help_text()
    
    def get_form_attrs(self) -> dict:
        attrs = super().get_form_attrs()

        size = getattr(self.object, f"{VARIABLES.PY_SIZE_VAR}_{self.field_name}", None)
        if not size:

            for hook in hooks.get_hooks(FIELD_EDITOR_SIZE):
                size = hook(self.object, self.meta_field)
                if size:
                    break
            
        if not size \
          and self.meta_field.is_relation \
          and use_related_form(self.meta_field):
            size = "full"
        
        if size:
            return attrs | {
                VARIABLES.FORM_SIZE_VAR: size,
            }
        
        return attrs

    def get_form(self):
        if self.request.method == "POST":
            form = self.form_class(self.request.POST, request=self.request, instance=self.object)
        else:
            form = self.form_class(request=self.request, instance=self.object)
        return form

    def form_valid(self, form):
        self.object = form.save()

        # Check if we are saving a relation
        if model_diff(self.object, self.original_object):
            self.meta_field.save_form_data(self.original_object, self.object)
            field_forms.save_possible_revision(self.original_object, self.request)

        extra_log_kwargs = {}
        if isinstance(self.original_object, RevisionMixin):
            extra_log_kwargs["revision"] = self.original_object.latest_revision

        with translation.override(None):
            old_new = {}
            if not isinstance(self.meta_field, StreamField):
                old_new = {
                    "old": str(self.initial_field_value),
                    "new": str(getattr(self.object, self.field_name)),
                }

            data = {
                "verbose_field_name": str(self.meta_field.verbose_name),
                "field_name": self.field_name,
                "model_id": self.object.pk,
                "model_name": self.object._meta.model_name,
                "app_label": self.object._meta.app_label,
                "model_verbose": str(self.model._meta.verbose_name),
                "model_string": str(get_model_string(self.original_object)),
                **old_new,
            }

            uid = uuid.uuid4()
            if self.original_object.pk != self.object.pk:
                data.update({
                    "edited_model_string": str(get_model_string(self.object)),
                    "edited_model_verbose": str(self.object._meta.verbose_name),
                    "edited_model_id": self.object.pk,
                    "edited_model_name": self.object._meta.model_name,
                    "edited_app_label": self.object._meta.app_label,
                })

                log(
                    instance=self.object,
                    action="wagtail_fedit.related_changed",
                    user=self.request.user,
                    uuid=uid,
                    data=data,
                    content_changed=True,
                )

            log(
                instance=self.original_object,
                action="wagtail_fedit.edit_field",
                user=self.request.user,
                uuid=uid,
                data=data,
                content_changed=True,
                **extra_log_kwargs,
            )

    def render_content(self, parent_context=None):
        return get_field_content(
            self.request,
            self.original_object,
            self.meta_field,
            parent_context,
        )



from .base import (
    Keyword,
    BaseAdapter,
    AdapterError,
)
from .registry import (
    AdapterRegistry,
    RegistryLookUpError,
    DuplicateAdapterError,
    AdapterSubclassError,
    registry as adapter_registry,
)
from .block import (
    BlockAdapter,
)
from .field import (
    FieldAdapter,
)
from .models import (
    ModelAdapter,
)
from .funcs import (
    FuncAdapterMixin,
    BaseFieldFuncAdapter,
    BaseBlockFuncAdapter,
)
from .misc import (
    BackgroundImageFieldAdapter,
)


from typing import Any, TYPE_CHECKING
from collections import namedtuple
from urllib.parse import urlencode
from django.db import models
from django.http import HttpRequest
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.template.loader import render_to_string
from django.conf import settings
from django.urls import reverse

from wagtail.models import (
    DraftStateMixin,
    WorkflowMixin,
    LockableMixin,
    PreviewableMixin
)
from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.blocks.stream_block import StreamValue
from wagtail.blocks.list_block import ListValue
from wagtail import (
    hooks,
    blocks,
    VERSION as WAGTAIL_VERSION,
)

from .toolbar import (
    FeditAdapterComponent,
    FeditAdapterEditButton,
)
from .hooks import (
    EXCLUDE_FROM_RELATED_FORMS,
    REGISTER_TYPE_RENDERER,
    REGISTER_FIELD_RENDERER,
    CONSTRUCT_ADAPTER_TOOLBAR,
)


if TYPE_CHECKING:
    from .adapters.base import BaseAdapter


TEMPLATE_TAG_NAME = "fedit"
FEDIT_PREVIEW_VAR = "_wagtail_fedit_preview"
USERBAR_MODEL_VAR = "_wagtail_fedit_userbar_model"
LOG_ACTION_TEMPLATES_AVAILABLE = WAGTAIL_VERSION < (6, 1, 0)


class FeditPermissionCheck:
    @staticmethod
    def has_perms(request: HttpRequest, model: Any) -> bool:

        user = request
        if isinstance(request, HttpRequest):
            user = request.user
        
        if (
            not user.is_authenticated\
            or not user.has_perm("wagtailadmin.access_admin")\
            or not user.has_perm(f"{model._meta.app_label}.change_{model._meta.model_name}")    
        ):
            return False
        
        return True


class FeditIFrameMixin:
    ERROR_TITLE = _("Validation Errors")

    HEADING_SUPPORTS_DRAFTS = _("Publishing Required")
    TITLE_SUPPORTS_DRAFTS = _("The object you are editing supports drafts.")
    TEXT_PUBLISH_DRAFTS = _("You must publish %(model)s to make any changes visible.")

    HEADING_NO_DRAFTS = _("No Publishing Required")
    TITLE_NO_DRAFTS = _("The object you are editing does not support drafts.")
    TEXT_NO_DRAFTS = _("You are not required to publish this object to make this change visible.")


    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["help_text"] = self.get_help_text()
        context["error_title"] = self.get_error_title()
        return context
    
    def get_error_title(self) -> str:
        return self.ERROR_TITLE
    
        
    def get_header_title(self) -> str:
        return _("Edit %(type)s '%(model)s'") % {
            "type": self.object._meta.verbose_name,
            "model": get_model_string(getattr(self, "absolute_instance", self.object), request=self.request),
        }

    def get_help_text(self) -> str:
        # No relations. Maybe draft support.
        if is_draft_capable(getattr(self, "absolute_instance", self.object)):
            return {
                "status": "warning",
                "heading": self.HEADING_SUPPORTS_DRAFTS,
                "title": self.TITLE_SUPPORTS_DRAFTS,
                "text": mark_safe(self.TEXT_PUBLISH_DRAFTS % {
                    "model": get_model_string(getattr(self, "absolute_instance", self.object), publish_url=True, request=self.request),
                })
            }
        
        return {
            "status": "info",
            "heading": self.HEADING_NO_DRAFTS,
            "title": self.TITLE_NO_DRAFTS,
            "text": mark_safe(self.TEXT_NO_DRAFTS)
        }


def use_related_form(field: models.Field) -> bool:
    """
    Check if a field should be included in the related forms.
    Internally used to make sure we use widgets instead
    of rendering a form for a Page, Image or Document model.
    """
    for hook in hooks.get_hooks(EXCLUDE_FROM_RELATED_FORMS):
        if hook(field):
            return False
    return True


def access_userbar_model(request: HttpRequest) -> models.Model:
    """
    Retrieve the model set for the userbar in the request.
    """
    return getattr(request, USERBAR_MODEL_VAR, None)

def with_userbar_model(request: HttpRequest, model: models.Model) -> HttpRequest:
    """
    Set the model to be available in the userbar.
    The request is shared easily between these contexts - might as well use it.
    """
    setattr(request, USERBAR_MODEL_VAR, model)
    return request

def get_block_name(block):
    """
    Return the block's type name for a block in a StreamField or ListBlock.
    eg. "heading", "paragraph", "image", "item", "column", etc.
    """
    if isinstance(block, StreamValue.StreamChild):
        return block.block_type
    elif isinstance(block, ListValue.ListChild):
        return "item"
    elif isinstance(block, ListValue):
        return block.block.name
    elif isinstance(block, (blocks.StructValue, blocks.BoundBlock)):
        return block.block.name
    else:
        raise ValueError("Unknown block type: %s" % type(block))
    
def get_block_path(block):
    """
    Get the current path part of a block in a StreamField or ListBlock.
    """
    if isinstance(block, StreamValue.StreamChild):
        return block.id
    elif isinstance(block, ListValue.ListChild):
        return block.id
    elif isinstance(block, ListValue):
        return block.block.name
    elif isinstance(block, (blocks.StructValue, blocks.BoundBlock)):
        return block.block.name
    else:
        raise ValueError("Unknown block type: %s" % type(block))

def find_block(block_id, field, contentpath=None):
    """
    Find a block in a StreamField or ListBlock by its ID.
    """
    if contentpath is None:
        contentpath = []

    # Check and cast field to iterable if necessary, but do not append non-StreamValue names to contentpath here.
    if not isinstance(field, StreamValue) and not hasattr(field, "__iter__"):
        field = [field]

    # Adjust for ListValue to get the iterable bound_blocks.
    if isinstance(field, ListValue):
        field = field.bound_blocks

    for block in field:
        # Determine the block's name only if needed to avoid premature addition to contentpath.
        block_name = get_block_path(block)
        
        if getattr(block, "id", None) == block_id:
            # Append the block name here as it directly leads to the target.
            return block, contentpath + [block_name]
        
        # Prepare to check children without altering the current path yet.
        if isinstance(block.value, blocks.StructValue):
            for _, value in block.value.bound_blocks.items():
                found, found_path = find_block(block_id, value, contentpath + [block_name])
                if found:
                    return found, found_path

        elif isinstance(block.value, (StreamValue, StreamValue.StreamChild, ListValue)):
            found, found_path = find_block(block_id, block.value, contentpath + [block_name])
            if found:
                return found, found_path

    # Return None and the current path if no block is found at this level.
    return None, contentpath



_renderer_map = {}
_field_renderer_map = {}
_looked_for_renderers = False


def _look_for_renderers():
    global _looked_for_renderers
    if not _looked_for_renderers:
        for hook in hooks.get_hooks(REGISTER_TYPE_RENDERER):
            hook(_renderer_map)

        for hook in hooks.get_hooks(REGISTER_FIELD_RENDERER):
            hook(_field_renderer_map)

        _looked_for_renderers = True


def get_field_content(request, instance, meta_field: models.Field, context, content=None):
    """
    Return the content for a field on a model.
    Also checks the model for a rendering method.
    The method should be named `render_fedit_{field_name}`.
    We wil also check for any hooks which may convert the content.
    """
    _look_for_renderers()

    if isinstance(meta_field, str):
        meta_field = instance._meta.get_field(meta_field)

    if hasattr(context, "flatten"):
        context = context.flatten()

    if not content:
        # Check for a rendering method if it exists
        if hasattr(instance, f"render_fedit_{meta_field.name}"):
            content = getattr(instance, f"render_fedit_{meta_field.name}")(request, context=context)
        else:
            content = getattr(instance, meta_field.name)

    for k, v in _field_renderer_map.items():
        if isinstance(meta_field, k):
            content = v(request, context, instance, content)
            break

    for k, v in _renderer_map.items():
        if isinstance(content, k):
            content = v(request, context, instance, content)
            break

    # The content might be a streamblock etc, we can render it as a block
    # if isinstance(content, (blocks.BoundBlock, blocks.StructValue)):
    if hasattr(content, "render_as_block"):
        content = content.render_as_block(context)

    # The content might otherwise have a render method.
    elif hasattr(content, "render"):
        content = content.render(context)

    return content

def is_draft_capable(model):
    """
    Check if a model is capable of drafts.
    """
    return isinstance(model, DraftStateMixin)\
        or type(model) == type\
        and issubclass(model, DraftStateMixin)

def model_diff(m1, m2):
    """
    Check if two model instances are different based on their type and primary key.
    Does not check for differences in the model's fields.
    This is used to determine if a relation is being saved.
    """
    return not (
        m1._meta.app_label == m2._meta.app_label
        and m1._meta.model_name == m2._meta.model_name\
        and m1.pk == m2.pk
    )


def edit_url(instance: models.Model, request: HttpRequest, hash = None, **params) -> str:
    """
        Return the edit URL for a given object and user (or request instead of user.)
        If none exists and the model is an instance of PreviewableMixin;
        return the wagtail_fedit:editable url; else an empty string.
    """

    user = request.user
    finder = AdminURLFinder(user)
    admin_url = finder.get_edit_url(instance)

    if not admin_url:

        # Check if the instance is a PreviewableMixin
        # and the user has permission to edit it.
        if isinstance(instance, PreviewableMixin)\
                and _can_edit(request, instance):
            
            admin_url = reverse(
                "wagtail_fedit:editable",
                args=[
                    instance.pk,
                    instance._meta.app_label,
                    instance._meta.model_name
                ],
            )
        else:
            return ""

    data = urlencode(params)
    if params:
        admin_url = f"{admin_url}?{data}"

    if hash:
        admin_url = f"{admin_url}#{hash}"
    
    return admin_url


def get_model_string(instance: models.Model, publish_url: bool = False, request: HttpRequest = None, target = "_blank", **params) -> str:
    """
    Get a string representation of a model instance. If the instance has a
    `get_admin_display_title` method, it will be used to get the string
    representation.
    If that method does not exist, the `title` attribute will be used.
    If the `title` attribute does not exist, the string representation of the
    instance will be used.
    If `publish_url` is True, the string will be wrapped in an anchor tag
    linking to the publish view for the instance.
    Permissions will not be checked. This is the responsibility of the caller.
    """
    model_string = getattr(instance, "get_admin_display_title", None)
    if model_string:
        model_string = model_string()
    else:
        model_string = getattr(instance, "title", str(instance))

    if publish_url:

        if is_draft_capable(instance):
            admin_url = reverse(
                "wagtail_fedit:publish",
                args=[
                    instance.pk,
                    instance._meta.app_label,
                    instance._meta.model_name
                ],
            )

            data = urlencode(params)
            if params:
                admin_url = f"{admin_url}?{data}"

        else:
            admin_url = edit_url(instance, request, **params)

        if admin_url:
            model_string = mark_safe(
                f'<a href="{admin_url}" target="{target}">{model_string}</a>'
            )

    return model_string

def _can_edit(request, obj: models.Model):
    """
    Check if the user has appropriate permissions to edit an object.
    Also requires the current request be on the `wagtail_fedit:editable` url
    (Or the preview variable to be True)
    """
    if not request or not obj:
        return False
    
    return (
        FeditPermissionCheck.has_perms(request, obj)\
        and getattr(request, FEDIT_PREVIEW_VAR, False)
    )

def user_can_publish(instance, user, check_for_changes: bool = True):
    """
    Check if a user can publish an object.
    Mostly comes from PagePermissionTester.can_publish
    """
    if not isinstance(instance, DraftStateMixin):
        return False
    
    if not instance.has_unpublished_changes\
        and check_for_changes:
        return False

    if hasattr(instance, "permission_policy"):
        return instance.permission_policy.user_has_permission(user, "publish")
    
    if not hasattr(instance, "permissions_for_user"):
        return False

    return instance.permissions_for_user(user).can_publish()

def user_can_unpublish(instance, user):
    """
    Check if a user can unpublish an object.
    Mostly comes from PagePermissionTester.can_unpublish
    """
    if not isinstance(instance, DraftStateMixin):
        return False
    
    if not hasattr(instance, "permissions_for_user"):
        return False
    
    return instance.permissions_for_user(user).can_unpublish()

def user_can_submit_for_moderation(instance, user, check_for_changes: bool = True):
    """
    Check if a user can submit an object for moderation.
    Mostly comes from PagePermissionTester.can_submit_for_moderation
    """
    if not getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
        return False

    if not instance.has_unpublished_changes\
        and check_for_changes:
        return False

    if not isinstance(instance, WorkflowMixin):
        return False
    
    if not hasattr(instance, "permissions_for_user"):
        return False
    
    return instance.permissions_for_user(user).can_submit_for_moderation()


_lock_info = namedtuple("lock_info", ["lock", "locked_for_user"])

def lock_info(object, user) -> _lock_info:
    """
        Returns the Lock instance (if any) and whether it is locked for the given user.
    """
    if isinstance(object, LockableMixin):
        lock = object.get_lock()
        locked_for_user = lock is not None and lock.for_user(
            user
        )
    else:
        lock = None
        locked_for_user = False

    return _lock_info(lock, locked_for_user)


def wrap_adapter(request: HttpRequest, adapter: "BaseAdapter", context: dict, run_context_processors: bool = False) -> str:
    if not context:
        context = {}

    items: list[FeditAdapterComponent] = [
        *adapter.get_toolbar_buttons(),
        FeditAdapterEditButton(request, adapter),
    ]

    for hook in hooks.get_hooks(CONSTRUCT_ADAPTER_TOOLBAR):
        hook(items=items, adapter=adapter)

    items = [item.render() for item in items]
    items = list(filter(None, items))

    reverse_kwargs = {
        "adapter_id": adapter.identifier,
        "app_label": adapter.object._meta.app_label,
        "model_name": adapter.object._meta.model_name,
        "model_id": adapter.object.pk,
    }

    if adapter.field_name is not None:
        reverse_kwargs["field_name"] = adapter.field_name

    shared = adapter.encode_shared_context()
    js_constructor = adapter.get_js_constructor()
    
    return render_to_string(
        "wagtail_fedit/content/editable_adapter.html",
        {
            "identifier": adapter.identifier,
            "adapter": adapter,
            "buttons": items,
            "shared": shared,
            "unique_id": adapter.get_element_id(),
            "js_constructor": js_constructor,
            "shared_context": adapter.kwargs,
            "parent_context": context,
            "edit_url": reverse(
                "wagtail_fedit:edit",
                kwargs=reverse_kwargs,
            ),
            "refetch_url": reverse(
                "wagtail_fedit:refetch",
                kwargs=reverse_kwargs,
            ),
        },
        request=request if run_context_processors else None,
    )



from django.db import models
from wagtail import hooks

import copy

from wagtail_fedit import (
    hooks as fedit_hooks,
    utils,
)

from .base import BaseFEditTest


class TestUtils(BaseFEditTest):
    def setUp(self):
        super().setUp()
        # Wrap output from textfield in paragraph.
        utils._field_renderer_map[models.TextField] =\
            lambda request, context, instance, value:\
                f"<p class=\"text-field\">{value}</p>"
        
        @hooks.register(fedit_hooks.EXCLUDE_FROM_RELATED_FORMS)
        def exclude_related_forms(field: models.Field):
            return field.related_model == self.full_model.__class__

    def test_get_field_content(self):
        request = self.request_factory.get("/")
        self.assertEqual(
            utils.get_field_content(request, self.basic_model, "title", {}),
            self.basic_model.title
        )
        self.assertEqual(
            utils.get_field_content(request, self.basic_model, "body", {}),
            f"<p class=\"text-field\">{self.basic_model.body}</p>"
        )

    def test_permission_check(self):
        self.assertTrue(
            utils.FeditPermissionCheck.has_perms(self.admin_user, self.basic_model)
        )
        self.assertFalse(
            utils.FeditPermissionCheck.has_perms(self.regular_user, self.basic_model)
        )
        self.assertFalse(
            utils.FeditPermissionCheck.has_perms(self.anonymous_user, self.basic_model)
        )

    def test_use_related_form(self):

        basic_field = models.ForeignKey(
            self.basic_model.__class__,
            on_delete=models.CASCADE,
        )

        full_field = models.ForeignKey(
            self.full_model.__class__,
            on_delete=models.CASCADE,
        )

        self.assertTrue(
            utils.use_related_form(basic_field)
        )
        self.assertFalse(
            utils.use_related_form(full_field)
        )

    def test_get_set_userbar_model(self):
        request = self.request_factory.get("/")

        self.assertIsNone(
            utils.access_userbar_model(request)
        )

        utils.with_userbar_model(request, self.basic_model)
        
        self.assertEqual(
            utils.access_userbar_model(request),
            self.basic_model
        )

    def test_is_draft_capable(self):
        self.assertTrue(utils.is_draft_capable(self.draft_model))
        self.assertFalse(utils.is_draft_capable(self.basic_model))

    def test_model_type_difference(self):
        self.assertTrue(
            utils.model_diff(self.full_model, self.basic_model)
        )
        self.assertFalse(
            utils.model_diff(self.full_model, self.full_model)
        )

        other_full_model = copy.deepcopy(self.full_model)
        other_full_model.title = "Other Title"

        self.assertFalse(
            utils.model_diff(self.full_model, other_full_model)
        )

    def test_can_edit(self):
        requests = (
            (self.admin_user, True),
            (self.regular_user, False),
            (self.anonymous_user, False),
        )

        for user, outcome in requests:
            request = self.request_factory.get("/")
            request.user = user

            self.assertFalse(
                utils._can_edit(request, self.basic_model)
            )

            setattr(
                request,
                utils.FEDIT_PREVIEW_VAR,
                True,
            )

            self.assertEqual(
                utils._can_edit(request, self.basic_model),
                outcome,
            )




from django.utils.translation import gettext_lazy as _
from django.http import HttpRequest

from wagtail.models import Page
from wagtail.admin.panels import (
    page_utils,
    model_utils,
    TabbedInterface,
)

from .base import (
    VARIABLES,
    Keyword,
    BlockFieldReplacementAdapter,
    AdapterError,
)
from ..forms import (
    PossibleRevisionFormMixin,
)
from ..utils import (
    get_model_string,
)
from wagtail.admin.admin_url_finder import (
    AdminURLFinder,
)
from ..toolbar import (
    FeditAdapterComponent,
    FeditAdapterAdminLinkButton,
)


class ModelAdapter(BlockFieldReplacementAdapter):
    """
    An adapter for editing any model.
    This will render the model and replace it on the frontend
    on successful form submission.
    """
    template_name = "wagtail_fedit/editor/adapter_edit_handler.html"
    identifier = "model"
    field_required = False
    usage_description = _("This adapter is used for directly editing a model instance.")
    keywords = BlockFieldReplacementAdapter.keywords + (
        Keyword(
            "render_method",
            optional=True,
            default="render_as_content",
            help_text="The method to call on the object to render it as content. Default is 'render_as_content'.",
        ),
    )

    def __init__(self, object, field_name: str, request: HttpRequest, **kwargs):
        super().__init__(object, field_name, request, **kwargs)
        if isinstance(self.object, Page):
            self.edit_handler = page_utils._get_page_edit_handler(self.object.__class__)
        else:
            self.edit_handler = model_utils.get_edit_handler(self.object.__class__)

    def get_admin_url(self) -> str:
        finder = AdminURLFinder(self.request.user)
        return finder.get_edit_url(self.object)

    def get_toolbar_buttons(self) -> list[FeditAdapterComponent]:
        buttons = super().get_toolbar_buttons()
        if not self.kwargs.get("admin", False):
            return buttons
        
        buttons.append(FeditAdapterAdminLinkButton(
            self.request, self,
        ))
        return buttons


    def get_form_attrs(self) -> dict:
        attrs = super().get_form_attrs()
        if isinstance(self.edit_handler, TabbedInterface):
            attrs[VARIABLES.FORM_SIZE_VAR] = "full"
        elif len(self.edit_handler.children) > 4:
            attrs[VARIABLES.FORM_SIZE_VAR] = "large"
        return attrs
    
    def get_form_context(self, **kwargs):
        context = super().get_form_context(**kwargs)
        bound_panel = self.edit_handler.get_bound_panel(
            instance=self.object, request=self.request, form=kwargs["form"],
        )
        context["edit_handler"] = bound_panel
        return context

    @property
    def form_class(self):
        cls = self.edit_handler.get_form_class()
        class RevisionModelForm(PossibleRevisionFormMixin, cls):
            pass
        return RevisionModelForm

    def form_valid(self, form):
        self.object = form.save()

    def get_form(self):
        if self.request.method == "POST":
            form = self.form_class(self.request.POST, for_user=self.request.user, instance=self.object, request=self.request)
        else:
            form = self.form_class(for_user=self.request.user, instance=self.object, request=self.request)
        return form

    def get_header_title(self):
        instance_string = get_model_string(self.object)
        return _("Edit model '%(instance_string)s'") % {
            "instance_string": instance_string,
        }
    
    def get_help_text(self):
        return None
          
    def get_element_id(self) -> str:
        m = self.model
        return f"model-{m._meta.app_label}-{m._meta.model_name}-{self.object.pk}"

    def form_valid(self, form):
        self.object = form.save()

    def render_content(self, parent_context=None):
        if hasattr(parent_context, "flatten"):
            parent_context = parent_context.flatten()

        method_name = self.kwargs["render_method"]
        if not hasattr(self.object, method_name):
            raise AdapterError(
                "Object '%s' does not have any method named '%s'" % (
                    self.object.__class__.__name__,
                    method_name,
                )
            )
        method = getattr(
            self.object,
            method_name
        )
        return method(
            request=self.request,
            context=parent_context,
        )



from django.http import HttpRequest
from django.utils.functional import classproperty
from .base import Keyword
from .field import (
    FieldAdapter,
)
from .funcs import (
    BaseFieldFuncAdapter,
)

from wagtail.images.models import Filter
from wagtail.images.utils import to_svg_safe_spec
from wagtail.images.shortcuts import (
    get_rendition_or_not_found,
)


class BackgroundImageFieldAdapter(BaseFieldFuncAdapter):
    """
    Adapter for background-images.
    On succesful form submission; will return the URL of the image.
    This is placed in the background-image property of the target element. 
    """
    inline = True
    identifier = "field_bg_image"
    usage_description = "This adapter is used for changing a css property of a target element to a background-image."
    keywords = (
        Keyword(
            "target",
            help_text="The target element to apply the background-image to - this should be a css selector.",
            type_hint="str",
        ),
        Keyword(
            "css_variable_name",
            optional=True,
            help_text="The CSS variable name to apply the background-image to. element.style.setProperty(css_variable_name, url);",
            type_hint="str",
        ),
        Keyword(
            "filter_spec",
            optional=True,
            default="original",
            help_text="The filter spec to apply to the image.",
        ),
        Keyword(
            "preserve_svg",
            optional=True,
            default=False,
            help_text="Preserve SVG images by converting them to a safe format.",
        ),
    )
    js_constructor = "wagtail_fedit.editors.WagtailFeditFuncEditor"
    js_function = "wagtail_fedit.funcs.backgroundImageFunc"

    def render_content(self, parent_context=None):
        return ""
    
    def get_response_data(self, parent_context=None):
        data = super().get_response_data(parent_context)
        image = getattr(self.object, self.field_name, None)
        if not image:
            return data

        filter_spec = self.kwargs["filter_spec"]
        if not isinstance(filter_spec, str):
            filter_spec = "|".join(filter_spec)

        if image.is_svg() or self.kwargs["preserve_svg"]:
            filter = Filter(to_svg_safe_spec(filter_spec.split("|")))
        else:
            filter = Filter(spec=filter_spec)

        rendition = get_rendition_or_not_found(
            image,
            filter,
        )

        return data | {
            "url": rendition.url,
            "css_variable_name":\
                self.kwargs.get("css_variable_name", None) or "background-image",
        }



from django.contrib.contenttypes.models import (
    ContentType,
)
from wagtail.log_actions import (
    registry,
    log,
)
from wagtail.models import (
    Page,
    Workflow,
    WorkflowTask,
    WorkflowContentType,
    GroupApprovalTask,
    ModelLogEntry,
)
from .base import (
    BaseFEditTest,
)
from ..models import (
    EditableFullModel,
)
from wagtail_fedit.views import (
    PublishView,
    SubmitView,
    UnpublishView,
    CancelView,
)
from wagtail_fedit.utils import (
    LOG_ACTION_TEMPLATES_AVAILABLE,
)

registry.register_model(EditableFullModel, ModelLogEntry)

class TestSubmitViews(BaseFEditTest):

    def makeRequest(self, url_name: str, model, view_class, check_status = 0):
        url = self.get_url_for(url_name,
            app_label=model._meta.app_label,
            model_name=model._meta.model_name,
            model_id=model.pk,
        )
        view = view_class()
        view.request = self.request_factory.post(url)
        expected_post_value = view.get_action_value()

        response = self.client.post(url, {"action": expected_post_value})
        if check_status:
            self.assertEqual(response.status_code, check_status, msg=f"Request failed for {url_name}")
        return response
    
    def test_get_publish(self):
        self.client.force_login(self.admin_user)

        for model in [self.page_model, self.full_model]:
    
            revision = model.save_revision(
                user=self.admin_user,
                clean=False,
            )

            log(
                instance=model,
                action="wagtail_fedit.edit_field",
                user=self.admin_user,
                revision=revision,
                content_changed=True,
            )

            log(
                instance=model,
                action="wagtail_fedit.edit_block",
                user=self.admin_user,
                revision=revision,
                content_changed=True,
            )

            url = self.get_url_for("publish",
                app_label=model._meta.app_label,
                model_name=model._meta.model_name,
                model_id=model.pk,
            )

            try:
                response = self.client.get(url)
                content = response.content.decode()

                self.assertEqual(response.status_code, 200)

                self.assertIn('class="wagtail-fedit-log-entries"', content)
                
                if LOG_ACTION_TEMPLATES_AVAILABLE:
                    if isinstance(model, Page):
                        self.assertIn('ul class="actions"', content) # Supported for pages
                    else:
                        self.assertNotIn('ul class="actions"', content) # Not supported for snippets / regular models

            except Exception as e:
                if isinstance(e, AssertionError):
                    self.fail(str(e))

                self.fail(f"Request failed for {url} with error: {e}")

    def test_get_submit(self):
        self.client.force_login(self.admin_user)

        url = self.get_url_for("submit",
            app_label=self.full_model._meta.app_label,
            model_name=self.full_model._meta.model_name,
            model_id=self.full_model.pk,
        )

        try:
            response = self.client.get(url)
            content = response.content.decode()
        except Exception as e:
            self.fail(f"Request failed for {url} with error: {e}")

        self.assertEqual(response.status_code, 200)

    def test_get_unpublish(self):
        self.client.force_login(self.admin_user)

        url = self.get_url_for("unpublish",
            app_label=self.full_model._meta.app_label,
            model_name=self.full_model._meta.model_name,
            model_id=self.full_model.pk,
        )

        try:
            response = self.client.get(url)
            content = response.content.decode()
        except Exception as e:
            self.fail(f"Request failed for {url} with error: {e}")

        self.assertEqual(response.status_code, 200)

    def test_get_cancel(self):
        self.client.force_login(self.admin_user)

        url = self.get_url_for("cancel",
            app_label=self.full_model._meta.app_label,
            model_name=self.full_model._meta.model_name,
            model_id=self.full_model.pk,
        )

        try:
            response = self.client.get(url)
            content = response.content.decode()
        except Exception as e:
            self.fail(f"Request failed for {url} with error: {e}")

        self.assertEqual(response.status_code, 200)

    def test_publish(self):
        self.client.force_login(self.admin_user)
        
        self.full_model.unpublish()

        self.assertFalse(self.full_model.live)

        self.makeRequest("publish", self.full_model, PublishView, 302)

        self.full_model.refresh_from_db()

        self.assertTrue(self.full_model.live)

    def test_unpublish(self):
        self.client.force_login(self.admin_user)

        revision = self.full_model.save_revision(
            user=self.admin_user,
            clean=False,
        )
        
        self.full_model.publish(
            user=self.admin_user,
            revision=revision,
        )

        self.assertTrue(self.full_model.live)

        self.makeRequest("unpublish", self.full_model, UnpublishView, 302)

        self.full_model.refresh_from_db()

        self.assertFalse(self.full_model.live)

    def test_submit(self):
        self.client.force_login(self.admin_user)
        
        self.full_model.has_unpublished_changes = True
        self.full_model.save()

        workflow = self.full_model.get_workflow()
        self.assertIsNone(workflow)

        workflow = Workflow.objects.get_or_create(name="test_workflow")[0]
        workflow_task = GroupApprovalTask.objects.create(name="test_task_1")
        content_type = ContentType.objects.get_for_model(self.full_model.__class__)
        WorkflowContentType.objects.create(content_type=content_type, workflow=workflow)
        WorkflowTask.objects.create(
            workflow=workflow, task=workflow_task, sort_order=1
        )

        wf = self.full_model.get_workflow()
        self.assertIsNotNone(wf)
        self.assertEqual(wf, workflow)
        self.assertFalse(not not self.full_model.workflow_states.all())

        self.makeRequest("submit", self.full_model, SubmitView, 302)

        self.full_model = self.full_model.__class__.objects.get(pk=self.full_model.pk)
        workflow: Workflow = self.full_model.get_workflow()
        current_state = self.full_model.current_workflow_state

        self.assertEqual(
            current_state.status,
            current_state.STATUS_IN_PROGRESS,
        )

    def test_cancel(self):
        self.test_submit()

        self.makeRequest("cancel", self.full_model, CancelView, 302)

        self.full_model.refresh_from_db()

        current_state = self.full_model.workflow_states.first()
        self.assertEqual(
            current_state.status,
            current_state.STATUS_CANCELLED,
        )

    def test_lock_nopublish(self):
        self.client.force_login(self.other_admin_user)
        
        self.lock_model.live = False
        self.lock_model.has_unpublished_changes = True
        self.lock_model.save(update_fields=["live", "has_unpublished_changes"])

        self.makeRequest("publish", self.lock_model, PublishView, 200)

        self.lock_model.refresh_from_db()

        self.assertFalse(self.lock_model.live)
        self.assertTrue(self.lock_model.has_unpublished_changes)

    def test_lock_nounpublish(self):
        self.client.force_login(self.admin_user)

        if not self.lock_model.latest_revision:
            self.lock_model.save_revision(
                user=self.admin_user,
                clean=False,
            )

        self.lock_model.publish(
            user=self.admin_user,
            revision=self.lock_model.latest_revision,
        )
        
        self.makeRequest("unpublish", self.lock_model, UnpublishView, 200)

        self.lock_model.refresh_from_db()

        self.assertTrue(self.lock_model.live)
        self.assertFalse(self.lock_model.has_unpublished_changes)

    def test_lock_nosubmit(self):
        self.client.force_login(self.admin_user)
        
        self.makeRequest("submit", self.lock_model, SubmitView, 200)

        self.lock_model.refresh_from_db()

        self.assertTrue(self.lock_model.has_unpublished_changes)
        self.assertFalse(not not self.lock_model.workflow_states)

    def test_lock_nocancel(self):
        self.lock_model.locked = False
        self.lock_model.locked_by = None
        self.lock_model.save(update_fields=["locked", "locked_by"])

        self.test_submit()

        self.lock_model.locked = True
        self.lock_model.locked_by = self.other_admin_user
        self.lock_model.save(update_fields=["locked", "locked_by"])
        
        self.makeRequest("cancel", self.lock_model, CancelView, 200)

        self.lock_model.refresh_from_db()
        current_state = self.full_model.current_workflow_state

        self.assertEqual(
            current_state.status,
            current_state.STATUS_IN_PROGRESS,
        )



# Generated by Django 5.0.4 on 2024-04-20 22:33

import django.db.models.deletion
import wagtail.blocks
import wagtail.fields
import wagtail.models
import wagtail_fedit.test.core.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0093_uploadedfile'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BasicModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
            ],
        ),
        migrations.CreateModel(
            name='EditablePageModel',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='EditablePreviewModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
            ],
            bases=(wagtail_fedit.test.core.models.BaseEditableMixin, wagtail.models.PreviewableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='EditableDraftModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('live', models.BooleanField(default=True, editable=False, verbose_name='live')),
                ('has_unpublished_changes', models.BooleanField(default=False, editable=False, verbose_name='has unpublished changes')),
                ('first_published_at', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='first published at')),
                ('last_published_at', models.DateTimeField(editable=False, null=True, verbose_name='last published at')),
                ('go_live_at', models.DateTimeField(blank=True, null=True, verbose_name='go live date/time')),
                ('expire_at', models.DateTimeField(blank=True, null=True, verbose_name='expiry date/time')),
                ('expired', models.BooleanField(default=False, editable=False, verbose_name='expired')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
                ('latest_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='latest revision')),
                ('live_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='live revision')),
            ],
            options={
                'abstract': False,
            },
            bases=(wagtail_fedit.test.core.models.BaseEditableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='EditableFullModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('live', models.BooleanField(default=True, editable=False, verbose_name='live')),
                ('has_unpublished_changes', models.BooleanField(default=False, editable=False, verbose_name='has unpublished changes')),
                ('first_published_at', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='first published at')),
                ('last_published_at', models.DateTimeField(editable=False, null=True, verbose_name='last published at')),
                ('go_live_at', models.DateTimeField(blank=True, null=True, verbose_name='go live date/time')),
                ('expire_at', models.DateTimeField(blank=True, null=True, verbose_name='expiry date/time')),
                ('expired', models.BooleanField(default=False, editable=False, verbose_name='expired')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
                ('latest_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='latest revision')),
                ('live_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='live revision')),
            ],
            options={
                'abstract': False,
            },
            bases=(wagtail_fedit.test.core.models.BaseEditableMixin, models.Model, wagtail.models.WorkflowMixin, wagtail.models.PreviewableMixin),
        ),
        migrations.CreateModel(
            name='EditableLockModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('live', models.BooleanField(default=True, editable=False, verbose_name='live')),
                ('has_unpublished_changes', models.BooleanField(default=False, editable=False, verbose_name='has unpublished changes')),
                ('first_published_at', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='first published at')),
                ('last_published_at', models.DateTimeField(editable=False, null=True, verbose_name='last published at')),
                ('go_live_at', models.DateTimeField(blank=True, null=True, verbose_name='go live date/time')),
                ('expire_at', models.DateTimeField(blank=True, null=True, verbose_name='expiry date/time')),
                ('expired', models.BooleanField(default=False, editable=False, verbose_name='expired')),
                ('locked', models.BooleanField(default=False, editable=False, verbose_name='locked')),
                ('locked_at', models.DateTimeField(editable=False, null=True, verbose_name='locked at')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
                ('latest_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='latest revision')),
                ('live_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='live revision')),
                ('locked_by', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='locked_%(class)ss', to=settings.AUTH_USER_MODEL, verbose_name='locked by')),
            ],
            options={
                'abstract': False,
            },
            bases=(wagtail_fedit.test.core.models.BaseEditableMixin, wagtail.models.WorkflowMixin, models.Model),
        ),
        migrations.CreateModel(
            name='EditableRevisionModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('content', wagtail.fields.StreamField([('heading_component', wagtail.blocks.StructBlock([('heading', wagtail.blocks.CharBlock(max_length=25)), ('subheading', wagtail.blocks.CharBlock(max_length=40))])), ('flat_menu_component', wagtail.blocks.StructBlock([('title', wagtail.blocks.CharBlock(max_length=25)), ('subtitle', wagtail.blocks.RichTextBlock()), ('items', wagtail.blocks.ListBlock(wagtail.blocks.StructBlock([('link', wagtail.blocks.StructBlock([('text', wagtail.blocks.CharBlock(max_length=25))]))])))]))])),
                ('latest_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.revision', verbose_name='latest revision')),
            ],
            options={
                'abstract': False,
            },
            bases=(wagtail_fedit.test.core.models.BaseEditableMixin, models.Model),
        ),
    ]



from typing import (
    TYPE_CHECKING, Any, Type,
)
from django import forms
from django.db import models
from django.utils import translation
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from django.utils.text import (
    slugify,
)
from django.core.signing import (
    Signer,
)
from django.http import (
    HttpRequest,
)
from ..settings import (
    SIGN_SHARED_CONTEXT,
    SHARE_WITH_SESSIONS,
    USE_ADAPTER_SESSION_ID,
    TRACK_LOCALES,
)
from ..utils import (
    FeditIFrameMixin,
    wrap_adapter,
)

import uuid

if TYPE_CHECKING:
    from ..toolbar import (
        FeditToolbarComponent,
    )

class AdapterError(Exception):
    pass


def content_id_from_parts(*parts: Any) -> str:
    return "-".join(map(slugify, map(str, parts)))


import pickle, json, base64



class VARIABLES:
    PY_SIZE_VAR = "editor_size"
    FORM_SIZE_VAR = "data-editor-size"


class PickleBlockSerializer:
    """
    Simple wrapper around pickle to be used in signing.dumps and
    signing.loads.
    """

    def dumps(self, obj):
        return pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)

    def loads(self, data):
        return pickle.loads(data)

def Base85_json_dumps(obj):
    return base64.b85encode(json.dumps(obj).encode("utf-8")).decode("utf-8")

def Base85_json_loads(data):
    return json.loads(base64.b85decode(data).decode("utf-8"))



class Keyword:
    def __init__(self, name: str, optional: bool = False, absolute: bool = False, default=None, help_text: str = None, type_hint: Type[Any] = None):

        if (default and absolute) or (default and not optional):
            raise AdapterError("Keywords cannot be absolute or required and have a default value")

        if optional and absolute:
            raise AdapterError("Keywords cannot be optional and absolute")
        
        if default is not None and not type_hint:
            type_hint = type(default)
        
        self.name = name
        self.optional = optional
        self.absolute = absolute
        self.default = default
        self.help_text = help_text
        self.type_hint = type_hint

    def __str__(self):
        k = []

        if self.optional:
            k.append("?")
        if self.absolute:
            k.append("!")

        k.append(self.name)

        if self.type_hint and not self.absolute:
            if isinstance(self.type_hint, str):
                k.append(f": {self.type_hint}")
            else:
                k.append(f": {self.type_hint.__name__}")
                
        if self.default is not None and not self.absolute:
            default = self.default
            if isinstance(default, str):
                default = f"'{default}'"
            k.append(f"={default}")

        return "".join(k)

    def __repr__(self):
        s = ["Keyword", self.name]
        if self.optional:
            s.append("optional")
        elif not self.absolute:
            s.append("required")
        if self.absolute:
            s.append("absolute")
        if self.default:
            s.append(f"default={self.default}")
        if self.help_text:
            s.append(f"help_text='{self.help_text}'")
        if self.type_hint:
            s.append(f"type: {self.type_hint}")
        s = " ".join(s)
        return f"<{s}>"

def _get_keywords(bases, attrs):
    if "keywords" in attrs:
        return list(attrs["keywords"])
    
    for base in bases:
        if hasattr(base, "keywords"):
            return list(base.keywords)
        
    return list()

def _sort_keywords(keywords):
    _s = sorted(keywords, key=lambda x: (not x.absolute, x.optional))
    return _s

class AdapterMeta(type):
    def __new__(cls, name, bases, attrs):
        keywords: list[Keyword] = _get_keywords(bases, attrs)
        required_kwargs = []
        absolute_tokens = []
        optional_kwargs = {}

        for keyword in keywords:
            if keyword.absolute:
                absolute_tokens.append(keyword.name)
            elif keyword.optional:
                if keyword.default is not None:
                    optional_kwargs[keyword.name] = keyword.default
                else:
                    optional_kwargs[keyword.name] = None
            else:
                required_kwargs.append(keyword.name)

        cls = super().__new__(cls, name, bases, attrs)

        cls.required_kwargs: tuple[str]     = tuple(required_kwargs)
        cls.absolute_tokens: tuple[str]     = tuple(absolute_tokens)
        cls.optional_kwargs: dict[str, Any] = optional_kwargs
        cls.keywords:        tuple[Keyword] = tuple(
            _sort_keywords(keywords)
        )

        return cls


class BaseAdapter(FeditIFrameMixin, metaclass=AdapterMeta):
    # How the adapter is identified on inside of the templatetag.
    identifier              = None

    # If the templatetag required the first argument to be model.field or just model
    field_required          = True

    # The template used to render the form.
    template_name           = "wagtail_fedit/editor/adapter_iframe.html"

    signer: Signer          = Signer()
    
    # Keyword arguments for the adapter
    keywords: tuple[Keyword]     = (
        Keyword(
            "inline",
            absolute=True,
            help_text="if passed; the adapter will be rendered with inline styles."
        ),
    )

    # An optional description to be displayed when running the help command.
    usage_description       = "This adapter is used to edit a field of a model instance."

    # The JS constructor for the adapter.
    # This will receive the data after a successful form submission.
    js_constructor          = None

    def __init__(self, object: models.Model, field_name: str, request: HttpRequest, **kwargs):
        self.object           = object
        self.request          = request
        self.kwargs           = kwargs

        for k, v in self.optional_kwargs.items():
            if k not in self.kwargs and v is not None:
                self.kwargs[k] = v

        if hasattr(request, "LANGUAGE_CODE") and TRACK_LOCALES:
            if "LANGUAGE_CODE" not in self.kwargs:
                self.kwargs["LANGUAGE_CODE"] = request.LANGUAGE_CODE

            if self.kwargs["LANGUAGE_CODE"] != request.LANGUAGE_CODE:
                translation.activate(self.kwargs["LANGUAGE_CODE"])

        if self.field_required:
            self.field_name     = field_name
            self.meta_field     = object._meta.get_field(field_name)
        else:
            self.field_name     = None
            self.meta_field     = None

    @classmethod
    def get_usage_string(cls) -> str:
        """
        Return a string which describes how to use the adapter.
        """
        s = []
        for keyword in cls.keywords:
            s.append(str(keyword))
                
        return " ".join(s)
    
    def get_form_context(self, **kwargs):
        return {
            "adapter": self,
            "field_name": self.field_name,
            "meta_field": self.meta_field,
        }

    @classmethod
    def get_usage_description(cls) -> str:
        """
        Return a description of how the adapter is used.
        """
        return cls.usage_description
    
    @classmethod
    def get_usage_help_text(cls) -> list[str]:
        """
        Return a help text which describes how to use the adapter.
        This might be a good time to exalain the kwargs.
        """
        d = {}
        for keyword in cls.keywords:
            if keyword.name not in d:
                d[keyword.name] = keyword.help_text
        return d
    
    def get_template_names(self) -> list[str]:
        """
        Return the template names for the adapter.
        """
        return [self.template_name]

    @property
    def field_value(self):
        """
        Call the value_from_object method on
        the meta field to get the value from the instance.
        """
        return getattr(self.object, self.field_name)
    
    @cached_property
    def model(self):
        """
        Return the model class of the object.
        """
        return self.object.__class__
    
    @cached_property
    def tooltip(self):
        """
        Return the tooltip for the adapter.
        """
        return self.get_header_title()
    
    def check_permissions(self):
        """
        Check if the user has the required permissions to edit the field.
        If false; the field will be rendered as normal (read only).
        """
        if not self.request.user.is_authenticated:
            return False
        return True
        
    def get_js_constructor(self) -> str:
        """
        Return the JS constructor for the adapter.
        This is used to link actions from the adapter to the frontend.
        """
        if not self.js_constructor:
            raise AdapterError("No JS constructor defined")
        
        return self.js_constructor

    def get_response_data(self, parent_context) -> dict:
        """
        The data which is returned to the frontend on a successful form submission.
        """
        return {
            "adapter": {
                "identifier":   self.identifier,
                "constructor":  self.get_js_constructor(),
                "element_id":   self.get_element_id(),
                "model": {
                    "pk":           self.object.pk,
                    "app_label":    self.model._meta.app_label,
                    "model_name":   self.model._meta.model_name,
                },
            }
        }
    
    def get_admin_url(self) -> str:
        """
        Return the admin URL for the object.
        """
        raise NotImplementedError

    def get_toolbar_buttons(self) -> list["FeditToolbarComponent"]:
        """
        Extra possible toolbar buttons.
        This is where for example; the edit icon goes.
        """
        return []
    
    def get_element_id_parts(self) -> list[str]:
        """
        Return the parts of the element ID.
        """
        return [
            self.model._meta.app_label,
            self.model._meta.model_name,
            self.object.pk,
            self.field_name,
        ]

    def get_element_id(self) -> str:
        """
        Return a unique identifier for the elements on the frontend.
        """
        return content_id_from_parts(
            *self.get_element_id_parts(),
        )
    
    def get_form_attrs(self) -> dict:
        """
        Extra possible form attributes rendered in the iFrame.
        """
        return {}
    
    def get_form(self) -> "forms.Form":
        """
        Return the form which knows how to handle this datatype.
        """
        raise NotImplementedError

    def form_valid(self, form: "forms.Form"):
        """
        Called if the form is valid; useful for saving the form or other things.
        """
        pass
    
    def form_invalid(self, form: "forms.Form"):
        """
        Called if the form is not valid.
        """
        pass

    @classmethod
    def render_from_kwargs(cls, context, **kwargs):
        """
        Render the content for the field from the kwargs if possible.
        This should NOT include the wagtail-fedit wrapper.
        This is rendered if anything goes wrong;
        like the model/field variables not being available.
        """
        raise AdapterError("Cannot render {} from kwargs".format(cls.__name__))

    def render_content(self, parent_context: dict = None) -> str:
        """
        Render the content for the field.
        This should NOT include the wagtail-fedit wrapper.
        """
        raise NotImplementedError
    
    def encode_shared_context(self) -> dict:
        """
        Encode a dictionary to a string.
        This will be passed as a GET parameter to the iFrame.
        Make sure the data is not too large.
        """
        if not self.kwargs:
            return ""
        
        if SHARE_WITH_SESSIONS:
            if USE_ADAPTER_SESSION_ID:
                # This ID should be uniquely generated for each adapter instance.
                # But this is not always possible - hence why we allow UUID's.
                id = self.get_element_id()
            else:
                # Set in kwargs, try not to clutter session data too much.
                id = self.kwargs.setdefault(
                    "wagtail_fedit_uuid",
                    str(uuid.uuid4())
                )
                
            self.request.session[id] = self.kwargs
            self.request.session.modified = True
            return id
        
        if SIGN_SHARED_CONTEXT:
            return self.signer.sign_object(self.kwargs, compress=True)
        
        return Base85_json_dumps(self.kwargs)


    @classmethod
    def decode_shared_context(cls, request: HttpRequest, object: models.Model, field: str, context: str) -> dict:
        """
        Decode an encoded contex string back to a dictionary.
        """
        if SHARE_WITH_SESSIONS:
            if not context:
                return {}
            return request.session.get(context, {})
        
        if SIGN_SHARED_CONTEXT:
            return cls.signer.unsign_object(context)# , serializer=PickleBlockSerializer)
        
        try:
            return Base85_json_loads(context)
        except json.JSONDecodeError:
            pass
        return {}
    
    
class BlockFieldReplacementAdapter(BaseAdapter):
    js_constructor = "wagtail_fedit.editors.BlockFieldEditor"

    def get_response_data(self, parent_context = None):
        data = super().get_response_data(parent_context)
        data["html"] = wrap_adapter(
            request=self.request,
            adapter=self,
            context=parent_context,
            run_context_processors=True
        )
        return data




from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .adapters import BaseAdapter

class FeditToolbarComponent:
    template_name = None
    permissions: list[str] = []

    def __init__(self, request):
        self.request = request

    def get_context_data(self):
        return {
            "self": self,
            "request": self.request,
        }
    
    def is_shown(self):
        if not all([self.request, self.request.user.is_authenticated]):
            return False
        
        if not self.permissions:
            return True
        
        return self.request.user.has_perms(self.permissions)
    
    def render(self):
        if not self.is_shown():
            return ""

        return mark_safe(render_to_string(
            self.template_name,
            self.get_context_data(),
        ))


class FeditAdapterComponent(FeditToolbarComponent):
    def __init__(self, request, adapter: "BaseAdapter"):
        super().__init__(request)
        self.adapter = adapter

class FeditAdapterEditButton(FeditAdapterComponent):
    """
        Required button class for the edit modal to function.
        This button is handled by the script in `wagtail_fedit/js/frontend.js`
    """
    template_name = "wagtail_fedit/content/buttons/edit_adapter.html"
    permissions = [
        "wagtailadmin.access_admin",
    ]


class FeditAdapterAdminLinkButton(FeditAdapterComponent):
    """
        Required button class for the edit modal to function.
        This button is handled by the script in `wagtail_fedit/js/frontend.js`
    """
    template_name = "wagtail_fedit/content/buttons/admin_link.html"
    permissions = [
        "wagtailadmin.access_admin",
    ]

    def __init__(self, request, adapter: "BaseAdapter"):
        super().__init__(request, adapter)
        self.url = adapter.get_admin_url()

    def is_shown(self):
        return super().is_shown() and bool(self.url)

    def get_context_data(self):
        return super().get_context_data() | {
            "admin_url": self.url,
        }
